#define _GNU_SOURCE
#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <errno.h>
#include <unistd.h>
#include <sys/ptrace.h>
#include <sys/prctl.h>
#include <sys/syscall.h>
#include <sys/signal.h>
#include <linux/seccomp.h>
#include <linux/filter.h>

int main(void) {
  // set some silly filter that blocks the chdir() syscall,
  // assuming that the syscall only makes x86-64 ABI syscalls.
  // not actually secure like this, just as an example...
  struct sock_filter filter[] = {
    BPF_STMT(BPF_LD | BPF_W | BPF_ABS, (offsetof(struct seccomp_data, nr))),
    BPF_JUMP(BPF_JMP | BPF_JEQ | BPF_K, __NR_chdir, 0, 1),
    BPF_STMT(BPF_RET | BPF_K, SECCOMP_RET_ERRNO | EUSERS),
    BPF_STMT(BPF_RET | BPF_K, SECCOMP_RET_ALLOW),
  };
  struct sock_fprog prog = { .len = sizeof(filter)/sizeof(filter[0]), .filter = filter };
  if (prctl(PR_SET_NO_NEW_PRIVS, 1, 0, 0, 0))
    err(1, "prctl");
  if (syscall(__NR_seccomp, SECCOMP_SET_MODE_FILTER, 0, &prog))
    err(1, "seccomp");

  errno = 0;
  printf("[parent] chdir(/) = %d (%m)\n", chdir("/"));

  pid_t child = fork();
  if (child == -1)
    err(1, "fork");
  if (child == 0) {
    prctl(PR_SET_PDEATHSIG, SIGKILL);
    errno = 0;
    printf("[child before sleep] chdir(/) = %d (%m)\n", chdir("/"));
    sleep(2);
    errno = 0;
    printf("[child after sleep] chdir(/) = %d (%m)\n", chdir("/"));
    raise(SIGSTOP);
    while (1)
      sleep(10);
  }

  sleep(1);
  printf("PTRACE_SEIZE with PTRACE_O_SUSPEND_SECCOMP\n");
  errno = 0;
  if (ptrace(PTRACE_SEIZE, child, NULL, PTRACE_O_SUSPEND_SECCOMP))
    err(1, "PTRACE_SEIZE");

  sleep(2);
  errno = 0;
  printf("setting PTRACE_O_SUSPEND_SECCOMP via PTRACE_SETOPTIONS: %d (%m)\n",
    ptrace(PTRACE_SETOPTIONS, child, NULL, PTRACE_O_SUSPEND_SECCOMP));
}
